package edu.princeton.swing.text;

import java.awt.*;
import java.awt.font.*;
import java.awt.geom.*;

/**
 * FontInfo is a class that analyzes a font to see if it is monospaced over a specified set of
 * characters.
 */
public class FontInfo
{
    private static final FontRenderContext DEFAULT_FONT_RENDER_CONTEXT =
        new FontRenderContext(null, false, false);

    private Font font;

    private int characterHeight;
    private int lineSpacing;

    private boolean monospaced;
    private int maxCharacterWidth;

    private int plainCharacterWidths[];
    private int boldCharacterWidths[];
    private int italicCharacterWidths[];
    private int boldItalicCharacterWidths[];


    /**
     * Creates a new FontInfo object.
     */
    public FontInfo(Font font, boolean characterAllowed[])
    {
        if (font == null || characterAllowed == null)
            throw new NullPointerException();

        Font plainFont      = font.deriveFont(Font.PLAIN);
        Font boldFont       = font.deriveFont(Font.BOLD);
        Font italicFont     = font.deriveFont(Font.ITALIC);
        Font boldItalicFont = font.deriveFont(Font.BOLD | Font.ITALIC);

        this.font = plainFont;

        // Create a test string
        int characterAllowedLength = characterAllowed.length;
        StringBuffer buffer = new StringBuffer(characterAllowedLength);

        for (int ctr = 0; ctr < characterAllowedLength; ctr++)
        {
            if (characterAllowed[ctr])
                buffer.append((char)ctr);
        }
        String testString = buffer.toString();

        // Calculate the maximum height and lineSpacing from the test string

        // The plain flavor's height
        LineMetrics lineMetrics = plainFont.getLineMetrics(testString, DEFAULT_FONT_RENDER_CONTEXT);
        characterHeight = (int)Math.ceil(lineMetrics.getAscent());
        lineSpacing     = (int)Math.ceil(lineMetrics.getDescent() + lineMetrics.getLeading());

        // The bold flavor's height
        lineMetrics = boldFont.getLineMetrics(testString, DEFAULT_FONT_RENDER_CONTEXT);
        characterHeight = Math.max(characterHeight, (int)Math.ceil(lineMetrics.getAscent()));
        lineSpacing     = Math.max(
            lineSpacing,
            (int)Math.ceil(lineMetrics.getDescent() + lineMetrics.getLeading())
            );

        // The italic flavor's height
        lineMetrics = italicFont.getLineMetrics(testString, DEFAULT_FONT_RENDER_CONTEXT);
        characterHeight = Math.max(characterHeight, (int)Math.ceil(lineMetrics.getAscent()));
        lineSpacing     = Math.max(
            lineSpacing,
            (int)Math.ceil(lineMetrics.getDescent() + lineMetrics.getLeading())
            );

        // The bold-italic flavor's height
        lineMetrics = boldItalicFont.getLineMetrics(testString, DEFAULT_FONT_RENDER_CONTEXT);
        characterHeight = Math.max(characterHeight, (int)Math.ceil(lineMetrics.getAscent()));
        lineSpacing     = Math.max(
            lineSpacing,
            (int)Math.ceil(lineMetrics.getDescent() + lineMetrics.getLeading())
            );


        // Calcualte the widths of the characters
        boolean monospaced = true;
        int maxCharacterWidth = 0;
        int plainCharacterWidths[]      = new int[characterAllowedLength];
        int boldCharacterWidths[]       = new int[characterAllowedLength];
        int italicCharacterWidths[]     = new int[characterAllowedLength];
        int boldItalicCharacterWidths[] = new int[characterAllowedLength];
        char testArray[] = new char[1];

        for (int ctr = 0; ctr < characterAllowedLength; ctr++)
        {
            if (characterAllowed[ctr])
            {
                testArray[0] = (char)ctr;

                // The plain flavor's width
                Rectangle2D stringBounds = plainFont.getStringBounds(
                    testArray, 0, 1,
                    DEFAULT_FONT_RENDER_CONTEXT
                    );

                int width = (int)Math.ceil(stringBounds.getWidth());
                plainCharacterWidths[ctr] = width;

                if (maxCharacterWidth == 0)
                {
                    maxCharacterWidth = width;
                }
                else if (width > maxCharacterWidth)
                {
                    maxCharacterWidth = width;
                    monospaced = false;
                }
                else if (width < maxCharacterWidth)
                {
                    monospaced = false;
                }

                // The bold flavor's width
                stringBounds = boldFont.getStringBounds(
                    testArray, 0, 1,
                    DEFAULT_FONT_RENDER_CONTEXT
                    );

                width = (int)Math.ceil(stringBounds.getWidth());
                boldCharacterWidths[ctr] = width;

                if (maxCharacterWidth == 0)
                {
                    maxCharacterWidth = width;
                }
                else if (width > maxCharacterWidth)
                {
                    maxCharacterWidth = width;
                    monospaced = false;
                }
                else if (width < maxCharacterWidth)
                {
                    monospaced = false;
                }

                // The italic flavor's width
                stringBounds = italicFont.getStringBounds(
                    testArray, 0, 1,
                    DEFAULT_FONT_RENDER_CONTEXT
                    );

                width = (int)Math.ceil(stringBounds.getWidth());
                italicCharacterWidths[ctr] = width;

                if (maxCharacterWidth == 0)
                {
                    maxCharacterWidth = width;
                }
                else if (width > maxCharacterWidth)
                {
                    maxCharacterWidth = width;
                    monospaced = false;
                }
                else if (width < maxCharacterWidth)
                {
                    monospaced = false;
                }

                // The bold-italic flavor's width
                stringBounds = boldItalicFont.getStringBounds(
                    testArray, 0, 1,
                    DEFAULT_FONT_RENDER_CONTEXT
                    );

                width = (int)Math.ceil(stringBounds.getWidth());
                boldItalicCharacterWidths[ctr] = width;

                if (maxCharacterWidth == 0)
                {
                    maxCharacterWidth = width;
                }
                else if (width > maxCharacterWidth)
                {
                    maxCharacterWidth = width;
                    monospaced = false;
                }
                else if (width < maxCharacterWidth)
                {
                    monospaced = false;
                }
            }
        }

        this.monospaced = monospaced;
        this.maxCharacterWidth = maxCharacterWidth;

        this.plainCharacterWidths      = plainCharacterWidths;
        this.boldCharacterWidths       = boldCharacterWidths;
        this.italicCharacterWidths     = italicCharacterWidths;
        this.boldItalicCharacterWidths = boldItalicCharacterWidths;
    }

    /**
     * Returns the font represented by this FontInfo object.
     *
     * @return The font represented by this FontInfo object.
     */
    public Font getFont()
    {
        return font;
    }

    /**
     * Returns wheter or not the font is monspaced over the specified set of characters.
     *
     * @return True iff the font is monospaced over the specified set of characters.
     */
    public boolean isMonospaced()
    {
        return monospaced;
    }

    /**
     * Returns the maximum width attained by one of the characters in the specified set of
     * characters.
     *
     * @return The maximum width attained by one of the characters in the specified set of
     * characters.
     */
    public int getMaxCharacterWidth()
    {
        return maxCharacterWidth;
    }

    /**
     * Returns the set of all character widths for the plain flavor of the font.
     *
     * @return The set of all character widths for the plain flavor of the font.  Indices which
     * are out of bounds and elements with a value 0 denote characters which were not in the
     * specified set.
     */
    public int[] getPlainCharacterWidths()
    {
        return plainCharacterWidths;
    }

    /**
     * Returns the set of all character widths for the bold flavor of the font.
     *
     * @return The set of all character widths for the bold flavor of the font.  Indices which
     * are out of bounds and elements with a value 0 denote characters which were not in the
     * specified set.
     */
    public int[] getBoldCharacterWidths()
    {
        return boldCharacterWidths;
    }

    /**
     * Returns the set of all character widths for the italic flavor of the font.
     *
     * @return The set of all character widths for the italic flavor of the font.  Indices which
     * are out of bounds and elements with a value 0 denote characters which were not in the
     * specified set.
     */
    public int[] getItalicCharacterWidths()
    {
        return italicCharacterWidths;
    }

    /**
     * Returns the set of all character widths for the bold-italic flavor of the font.
     *
     * @return The set of all character widths for the bold-italic flavor of the font.  Indices
     * which are out of bounds and elements with a value 0 denote characters which were not in the
     * specified set.
     */
    public int[] getBoldItalicCharacterWidths()
    {
        return boldItalicCharacterWidths;
    }

    /**
     * Gets the maximum character height for the font over the specified set of characters.
     *
     * @return The maximum character height for the font over the specified set of characters.
     */
    public int getCharacterHeight()
    {
        return characterHeight;
    }

    /**
     * Gets the recommended number pixels which should vertically separate each line.
     *
     * @return The recommended number pixels which should vertically separate each line.
     */
    public int getLineSpacing()
    {
        return lineSpacing;
    }
}
